#include <setjmp.h>
#define FSKFUNC __cdecl
#pragma comment(lib,"fsk.lib")
#define DWORD unsigned int
#define WORD unsigned short
#define BYTE unsigned char
#ifndef __cplusplus
#define bool unsigned char
#define true 1
#define false 0
#define FskInt64 __int64
#endif

//DocumentViewer stuff

enum propType {
  PropType_Integer = 1,
  PropType_Float,
  PropType_Boolean,
  PropType_String,
  PropType_Rect,
  PropType_Point,
  PropType_Pointer,
  PropType_HostObject,
  PropType_Duration,
  PropType_Dimensions = 15,
  PropType_Array = 0x10001,
  PropType_Method = 0x10002,
};

enum docViewerProps {
  DocViewerProp_dimensions=1,        //dimension
  DocViewerProp_pages,               //int
  DocViewerProp_page,                //int
  DocViewerProp_textScale,           //int
  DocViewerProp_crop,                //rect
  DocViewerProp_contents,            
  DocViewerProp_flow,                //boolean
  DocViewerProp_links,
  DocViewerProp_DRMInfo,
  DocViewerProp_DRMKey,
  DocViewerProp_layoutCache,         //pointer
  DocViewerProp_part,
  DocViewerProp_sourceDimensions,    //dimension
  DocViewerProp_font,                //string
  DocViewerProp_onProgress,          //method
  DocViewerProp_textEngine,          //string
  DocViewerProp_canCopy,
};

typedef struct {
  int x, y, width, height;
} FskRectangle;

typedef struct {
  int x, y;
} FskPoint;

typedef struct {
  int width, height;
} FskDimension;

typedef struct {
   int type;
   int id; //?
   union {
   	int iValue;
   	float fValue;
   	bool bValue;
   	char* strValue;
   	FskRectangle rectValue;
   	FskPoint ptValue;
   	struct { void* ptr; int size;} ptrValue;
   	void* hostObjValue;
   	int durationValue;
   	FskDimension dimValue;
    struct {void* func; void* instance;} valCustomMethod;
   };
} PropertyValue;

typedef int (*PropHandler)(void* pInstance, int unknown1, int unknown2, PropertyValue* prop);

typedef struct {
  int id;
  int type;
  //void* (*getter)(void* pInstance, int unknown1, int unknown2, PropertyValue* prop);
  //void* (*setter)(void* pInstance, int unknown1, int unknown2, PropertyValue* prop);
  PropHandler getter;
  PropHandler setter;
} PropertyDesc;

typedef struct _FskBitmap FskBitmap;
typedef struct _FskPort FskPort;
typedef struct _FskWindow FskWindow;
typedef struct _FskHover FskHover;
typedef struct _FskWindowExclude FskWindowExclude;
typedef struct _FskTextFormat FskTextFormat;

typedef struct {
  BYTE r,g,b,a;
} FskColor;

struct _DocViewer;

typedef struct
{
  /* 0x0/0 */   const char* pszMimeType;
  /* 0x4/4 */   struct _DocViewer* pCallTable;
  /* 0x8/8 */   void* pViewerInstance;
  /* 0xC/12 */  int (*fileReadProc)(void *data, DWORD nNumberOfBytesToRead, void* buffer, DWORD* NumberOfBytesRead, DWORD offset);
  /* 0x10/16 */ void * fileReadProcData;
  /* 0x14/20 */ DWORD padding;
  /* 0x18/24 */ FskInt64 nFilesize;
  /* 0x20/32 */ const char* pszFileUrl;
  /* 0x24/36 */ DWORD unk;
} ViewerData;

#define MARK_PAGESTART 1
#define MARK_PAGEEND   2

//all functions are cdecl (caller purged)
typedef struct _DocViewer {
 /* 0*/ int (*CanHandle)(const char* mimeType);                            //1 = can handle
 /* 4*/ int (*New)(ViewerData* hostData, void** ppInstance);                     //0 = success
 /* 8*/ int (*Dispose)(ViewerData* hostData, void* pInstance);                      //0 = success
 /* C*/ int (*Render)(ViewerData* hostData, void* pInstance, FskBitmap** ppBitmap); //0=success
 /*10*/ void* Find;
 /*14*/ void* GetMetadata;
 /*18*/ int (*MarkNewFromLocation)(ViewerData* hostData, void* pInstance, void **pMark, int pageNo, FskPoint *point, char flag);
 /*1C*/ int (*MarkDispose)(ViewerData* hostData, void* pInstance, void *mark);
 /*20*/ int (*MarkGetLocation)(ViewerData* hostData, void* pInstance, void *mark, int *pageNo, FskPoint *point, int *part);
 /*24*/ int (*MarkCompare)(ViewerData* hostData, void* pInstance, void* mark1, void* mark2, int* result);
 /*28*/ int (*SpanGetBounds)(ViewerData* hostData, void* pInstance, void *marks[2], int *rectCount, FskRectangle **pRects);
 /*2C*/ int (*SpanGetText)(ViewerData* hostData, void* pInstance, void *marks[2], char **pText);
 /*30*/ PropertyDesc* PropTable; //ends with an all-zero entry
 /*34*/ void* MarkSerialize;
 /*38*/ void* MarkParse;
} DocViewer;

#ifdef __cplusplus
extern "C" {
#endif

int FskMemPtrNew(int size, void** pptr); //0=OK
int FskMemPtrNewClear(int size, void** pptr); //0=OK
void* FskMemPtrAlloc(int size);
void* FskMemPtrCalloc(int size);
void FskMemPtrDispose(void* ptr);
void FskMemPtrDisposeAt(void** pptr);
int FskMemPtrNewFromData(int size, void* src, void** pptr);
int FskMemPtrRealloc(int size, void** pptr);
int FskMemPtrReallocC(int size, void** pptr);
void FskMemSet(void* ptr, char fill, int size);
int FskMemCompare(void* dest, void* src, int len);
int FskMemCopy(void* dest, void* src, int len);
void* FskMemMove(void* dest, void* src, int len);

#define FSK_NEWPTR(type, ptr) (FskMemPtrNewClear(sizeof(type), ptr))
#define FSK_NEW(type) ((type*)FskMemPtrAlloc(sizeof(type)))
#define FSK_DELETE(ptr) FskMemPtrDispose(ptr)

void* FskBSearch(const void *key, const void *base, size_t num, size_t width, int ( __cdecl *compare ) ( const void *, const void *) );

char* FskStrDoCopy(const char* src);
int FskStrLen(const char* src);
int FskUnicodeStrLen(wchar_t *str);
char* FskStrDoCat(const char* src1, const char* src2); //returns a new pointer
void FskStrCopy(char* dst, const char* src);
void FskStrCat(char* appendTo, const char* appendFrom);
void FskStrNCopy(char* dst, const char* src, size_t size);
void FskStrNCat(char* appendTo, const char* appendFrom, size_t size);
int FskStrCompare(const char* str1, const char* str2);
int FskStrCompareWithLength(const char* str1, const char* str2, size_t len);
int FskStrCompareCaseInsensitive(const char* str1, const char* str2);
int FskStrCompareCaseInsensitiveWithLength(const char* str1, const char* str2, size_t len);
char* FskStrStr(const char* str1, const char* str2);
char* FskStrChr(const char* str, char ch);
char* FskStrRChr(const char* str, char ch);
void FskStrNumToStr(int num, char* buf);
int FskStrToNum(const char* str);
char* FskStrStripHeadSpace(char* str);
FskInt64 FskStrToFskInt64Base(char* str, int base); //base<=16
FskInt64 FskStrToFskInt64(char* str);
char* FskStrNumToHex(int num, char* buf, int len);
int FskStrToNum(const char *);
int FskStrToL(const char *,char **,int);
unsigned int FskStrToUL(const char *,char **,int);
double FskStrToD(const char *,char **);
int FskStrTail(const char *, const char *);//looks like strcmp from the end?
char* FskStrCopyUntil(char *dst, const char *src, char stopchar);
char* FskStrNCopyUntil(char *dst, const char *src, int len,char stopchar);
char* FskStrNChr(const char *str, int len, char ch);
char* FskStrStripHeadSpace(char *str); //skip spaces and tabs
char* FskStrStripTailSpace(char *str);
int FskStrHexToNum(const char *str, int len);
void FskStrDecodeEscapedChars(const char *str1,char *str2);
int FskStrURLEncode(const char *str1,char *str2);
int FskStrListLen(const char* strlist);
int FskStrListDoCopy(char *src,char **pDst);
int FskStrListCompare(const char* list1, const char* list2);

#define EXTENSION_DOCVIEWER 11
#define EXTENSION_TEXTENGINE 15
void* FskExtensionGetByIndex(int type, int index);
int FskExtensionGetCount(int type);
void FskExtensionInstall(int type, void* desc);
void FskExtensionUninstall(int type, void* desc);

typedef struct {
  int tmAscent;             //part of the font above baseline
  int tmDescent;            //part of the font below baseline
  int tmExternalLeading;    //extra spacing to be added between lines of text. N.B.: for some reason FreeType returns negative value
} FskFontInfo;

//for FskPortSetTextAlignment
#define TA_CENTER 0
#define TA_LEFT   1
#define TA_RIGHT  2
#define TA_TOP    3
#define TA_BOTTOM 4
#define TA_SCALE  5

//for FskPortSetTextStyle
#define TS_PLAIN           0
#define TS_BOLD            1
#define TS_ITALIC          2
#define TS_UNDERLINE       4
#define TS_STRIKEOUT       0x10
#define TS_DOUBLEGRAYSCALE 0x20   //???
#define TS_TRUNCATE_END    0x100
#define TS_TRUNCATE_CENTER 0x200

//for FskPortTextFitWidth
#define FW_WORDS 4 //try to fit only whole words

//style values for FskWindowNew
#define FWS_FULLSCREEN 8
#define FWS_CUSTOMWIN  0x10

//for FskWindowSetTransition
#define WT_SLIDEUP   1
#define WT_PUSHUP    2
#define WT_ZOOMUP    3
#define WT_CROSSFADE 4
#define WT_STEPFADE  5
#define WT_PUSHLEFT  6
#define WT_PUSHRIGHT 7

typedef struct {
	int size;         //sizeof(FskGraphMode2)
	int nBlendLevel;
} FskGraphMode2;

#define GM1_BILINEARINT 0x1000000
#define GM1_COPY        0x0000001
#define GM1_ALPHA       0x0000002
#define GM1_COLORIZE    0x0000003

int FskWindowNew(FskWindow **pWindow, int width, int height, int style, int handler, int handlerData);
int FskWindowDispose(FskWindow *window);
int FskWindowGetSize(FskWindow *win,int *w,int *h);
int FskWindowSetSize(FskWindow *win,int width,int height);
int FskWindowEventSetHandler(FskWindow *window,int handler,int handlerData);
int FskWindowEventGetHandler(FskWindow *window,int* handler,int* handlerData);
int FskWindowSetLocation(FskWindow *window,int X,int Y);
int FskWindowGetLocation(FskWindow *window,int *X,int *Y);
int FskWindowSetTitle(FskWindow *window, char *title);
int FskWindowGetTitle(FskWindow *window, char **title);
int FskWindowSetVisible(FskWindow *window, bool visible);
int FskWindowGetVisible(FskWindow *window, bool *visible);
int FskWindowSetSizeConstraints(FskWindow *window, int x, int y, int w, int h);
int FskWindowGetSizeConstraints(FskWindow *window, int* x, int* y, int* w, int* h);
FskPort* FskWindowGetPort(FskWindow *window);
int FskWindowGetCursor(FskWindow *window, FskPoint *pos);
int FskWindowGetNativeWindow(FskWindow *window);//HWND on Win32
FskWindow * FskWindowGetActive();
int FskWindowUpdate(FskWindow *window);
int FskWindowSetTransition(FskWindow *window,int transitionType,int transitionDuration);
int FskWindowCopyBitsToWindow(FskWindow *window, FskBitmap *bitmap, FskRectangle *rectFrom, FskRectangle *rectTo);
void * FskWindowHoverAdd(FskWindow *window, FskRectangle *rect, FskHover **pHover);
int FskWindowHoverRemove(FskHover *hover);
int FskWindowHoverSetBounds(FskHover *hover, FskRectangle *bounds);
int FskWindowExcludeAdd(FskWindow *window, FskRectangle *bounds, FskWindowExclude **pExclude);
int FskWindowExcludeRemove(FskWindowExclude *exclude);	
int FskWindowExcludeSetBounds(FskWindowExclude *exclude, FskRectangle *bounds);

int FskPortNew(FskPort** pPort, FskBitmap* bitmap, const char* textEngine);
int FskPortDispose(FskPort* port);
int FskPortTextGetEngine(FskPort* port, char** pEngine);

int FskPortSetOrigin(FskPort *port, int X,int Y);
int FskPortGetOrigin(FskPort *port, int* pX,int* pY);
int FskPortSetUpdateRectangle(FskPort *port, FskRectangle *rect); //rect==0: full region
int FskPortGetUpdateRectangle(FskPort *port, FskRectangle *rect);
int FskPortSetClipRectangle(FskPort *port, FskRectangle *rect); //rect==0: full region
int FskPortGetClipRectangle(FskPort *port, FskRectangle *rect);
int FskPortSetPenColor(FskPort* port, FskColor* color);
int FskPortGetPenColor(FskPort* port, FskColor* color);
int FskPortSetTextAlignment(void* port, WORD horAlign, WORD verAlign);
int FskPortGetTextAlignment(void* port, WORD* horAlign, WORD* verAlign);
int FskPortSetTextSize(FskPort* port, WORD size); //size==0: default (12pt)
int FskPortGetTextSize(FskPort* port, WORD* size);
int FskPortSetTextStyle(void* port, int style);
int FskPortGetTextStyle(void* port, int* style);
int FskPortSetTextFont(FskPort* port, const char* fontname); //0=OK
int FskPortGetTextFont(FskPort* port, char** pFontname); //0=OK
int FskPortSetGraphicsMode(FskPort *port,int mode1, FskGraphMode2 *mode2);

int FskPortTextFormatNew(FskTextFormat **pFormat, FskPort *port, int fontSize, int style, char *szFontName);
int FskPortTextFormatDispose(FskTextFormat *format);
int FskPortTextFormatApply(FskTextFormat *format);

int FskPortResetInvalidRectangle(FskPort* port);
int FskPortInvalidateRectangle(FskPort* port, FskRectangle *rect);
int FskPortGetInvalidRectangle(FskPort* port, FskRectangle *rect);
int FskPortSetBitmap(FskPort* port, FskBitmap* bmp);
int FskPortGetBitmap(FskPort* port, FskBitmap** pBmp);
int FskPortGetBounds(FskPort* port, FskRectangle *rect);
int FskPortBeginDrawing(FskPort* port);
int FskPortEndDrawing(FskPort* port);
int FskPortAccumulateChange(FskPort* port, FskRectangle *rect);
int FskPortRectangleFill(FskPort* port, FskRectangle *rect);
int FskPortRectangleFrame(FskPort* port, FskRectangle *rect);
int FskPortRectangleGradient(FskPort *port, FskRectangle *rect, FskColor *color1,FskColor *color2);
//FskPortBitmapDraw 
//FskPortApplyMaskAndValue
//FskPortBitmapFill


//text is in UTF-8
int FskPortTextDraw(FskPort* port, const char* text, int len, const FskRectangle* r);
int FskPortStringDraw(FskPort* port, const char* text, const FskRectangle* r);
int FskPortTextGetBounds(FskPort* port, const char* text, int len, FskRectangle* r);
int FskPortStringGetBounds(FskPort* port, const char* text, FskRectangle* r);

//returns the maximum string position which still fits into maxWidth
int FskPortTextFitWidth(FskPort* port, char *str, int len, int maxWidth, char flag, int *bytepos, int *charpos);

int FskPortGetFontInfo(FskPort* port, FskFontInfo* info);

void FskRectangleIntersect(FskRectangle *rect1, FskRectangle *rect2, FskRectangle *result);
void FskRectangleUnion(FskRectangle *rect1, FskRectangle *rect2, FskRectangle *result);
bool FskRectangleIsEmpty(FskRectangle *rect);
bool FskRectangleIsEqual(FskRectangle *rect1, FskRectangle *rect2);
void FskRectangleOffset(FskRectangle *rect, int offX, int offY);
void FskRectangleInset(FskRectangle *rect, int X, int Y);
FskRectangle* FskRectangleSet(FskRectangle *rect, int X, int Y, int W, int H);
FskRectangle* FskRectangleSetFull(FskRectangle *rect);
FskRectangle* FskRectangleSetEmpty(FskRectangle *rect);
bool FskRectangleContainsPoint(FskRectangle *rect, FskPoint *pt);
FskRectangle* FskRectangleScaleToFit(FskRectangle *r1, FskRectangle *r2, FskRectangle *result);
bool FskRectanglesDoIntersect(FskRectangle *r1,FskRectangle *r2);
int FskRectangleFill(FskBitmap *bitmap, FskRectangle *rect, FskColor *color, int mode1, int mode2);

//type for FskBitmapNew
//1,5 = 24 bit
//2,3,4,16 = 32 bit
//6,7,8,15,17 = 16 bit (7=RGB565)
//11,12 = 8 bit gray
//10,9 = 8 bit?
//13 = 2 bit
//1,7,2,3,11 = real dib section

int FskBitmapNew(int width, int height, int type, FskBitmap **pBitmap);
int FskBitmapDispose(FskBitmap *bitmap);
int FskBitmapGetBits(FskBitmap *bitmap,int *pBits,int *lineLen,int *type);
int FskBitmapGetDepth(FskBitmap *bitmap, int* depth);
int FskBitmapGetBounds(FskBitmap *bitmap, FskRectangle *bounds);
int FskBitmapSetHasAlpha(FskBitmap *bitmap, bool hasAlpha);
int FskBitmapGetHasAlpha(FskBitmap *bitmap, bool* pHasAlpha);
int FskBitmapGetPixel(FskBitmap *bitmap, int x, int y, FskColor *value);
int FskBitmapDraw(FskBitmap *bmpFrom,FskRectangle *rectFrom,FskBitmap *bmpTo,FskRectangle *rectTo,int a5,int a6,int a7,int a8);
int FskTransferBitmap(FskBitmap *bmpFrom,FskRectangle *rectFrom,FskBitmap *bmpTo,FskRectangle *rectTo,int a5,int a6,int a7,int a8);

typedef struct _FskTextEngine FskTextEngine;
int FskTextEngineNew(FskTextEngine** pEngine, const char* engineName);
int FskTextEngineDispose(FskTextEngine* engine);
int FskTextAddFontFile(FskTextEngine* engine, const char* fontfile);
int FskTextBox(FskTextEngine* engine, FskBitmap* bitmap, const char* text, int len, FskRectangle* rect1, FskRectangle* rect2, FskColor* color, int, int textSize, int textStyle, int alignmentX, int alignmentY, char* pszFontName, void* formatCache);
//FskTextGetBounds

char* FskTextLatin1ToUTF8(char* src,int srclen, char** dst,int* dstlen);//dstlen can be zero
char* FskTextUnicode16BEToUTF8(wchar_t* src, int srclen, char** dst, int* dstlen); //lengths are in bytes
char* FskTextUnicode16LEToUTF8(wchar_t* src, int srclen, char** dst, int* dstlen); //lengths are in bytes
char* FskTextUnicode16SwapBytes(wchar_t* src, wchar_t* dsc, int len); //length is in wchars
//int FskTextUTF8ToUnicode16NE();
int FskTextUTF8StrLen(char* src); //return number of utf8 characters (not bytes)
char FskTextUTF8DefaultWhitespaceProcessing(char *str);
bool FskTextUTF8IsWhitespace(char* src, int srclen, int* spaceLen);//checks for space/tab/zero/fullwidth space
int FskTextUnicodeToUTF8Offset(char* src, int charno); //returns byte offset of a character number charno, -1 for failure
int FskTextUTF8Advance(char* str, int offset, int delta, int whence); //return how many bytes to advance for one UTF8 character. whence>0: one character forward, whence<0: one character backwards
int FskTextToUTF8(char *src,int srclen,char **dst,int *dstlen); //convert from local codepage to UTF8
int FskTextToPlatform(char *src,int srclen,char **dst,int *dstlen); //convert from UTF8 to local codepage

#ifdef __cplusplus
}
#endif
