#include "fsk.h"
#include <stdio.h>

//dynamic array of ints
typedef struct {
  int* m_data;
  int m_capacity;
  int m_size;
} IntArray;

void arraySetCapacity(IntArray* array, int capacity)
{
  if (array)
  {
    FskMemPtrRealloc(capacity*sizeof(int), &array->m_data);
    array->m_capacity = capacity;
  }
}

IntArray* arrayNew(int capacity)
{
  IntArray* arr = FSK_NEW(IntArray);
  arr->m_size = 0;
  arr->m_data = NULL;
  arraySetCapacity(arr, capacity);
  return arr;
}

void arrayDispose(IntArray* array)
{
  if (array) 
  {
    FSK_DELETE(array->m_data);
    FSK_DELETE(array);
  }
};

int arrayGetItem(IntArray* array, int idx)
{
  return (array && idx<array->m_size)?array->m_data[idx]:-1;
}

int arrayGetSize(IntArray* array)
{
  return (array)?array->m_size:-1;
}

void arraySetItem(IntArray* array, int idx, int value)
{
  if (!array) return;
  if (array->m_capacity<=idx)
    arraySetCapacity(array, idx+idx/2);
  if (array->m_size<=idx)
    array->m_size = idx+1;
  array->m_data[idx]=value;
}

void arrayAppendItem(IntArray* array, int value)
{
  if (array) arraySetItem(array, array->m_size, value);
}

typedef struct {
  ViewerData* pData;          //various info from the host, including file reading proc
  FskDimension m_dimension;   //dimensions of the output
  FskDimension m_srcDimension;// ditto?
  FskRectangle m_crop;        //cropping rectangle
  FskPort* m_port;            //graphics port for drawing text
  //bool m_bFlow;
  int m_scale;                //zoom value (0-2)
  short m_fontSize;           //current font size
  char* pszFontname;          //current font face name
  char* fileContents;         //file string (UTF-8)
  int fileSize;               //length of file string
  IntArray* m_arrParas;       //array of offsets to paragraph starts
  IntArray* m_arrLines;       //array of offsets to lines starts
  bool m_bDidParas;           //did we do paragraph detection already?
  bool m_bNeedToPaginate;     //do we need to redo line wrapping?
  int m_lineHeight;           //line height in pixels
  int m_linesPerPage;         //lines per page
  int m_curPage;              //current page number
  int m_totalPages;           //total page count
} myViewer;

int myCanHandle(const char* mimeType)
{
  //return FskStrCompare(mimeType,"text/blabla");
  printf("[testviewer] CanHandle(%s)\n",mimeType);
  if (!FskStrCompare(mimeType,"text/plain2"))
  	return 1;
  return 0;
}

int myNew(ViewerData* hostData, void** pViewer)
{
  myViewer* viewer;
  int err;
  int bytesRead;
  printf("[testviewer] New\n");
  printf("File: %s, type: %s, size: %d\n", hostData->pszFileUrl, hostData->pszMimeType, hostData->nFilesize);
  //FskMemPtrNew();
  if (!pViewer)
  	return 1;
  viewer = FSK_NEW(myViewer);
  *pViewer = viewer;
  if (!viewer)
    return 1;
  FskRectangleSetFull(&viewer->m_crop);
  err = FskPortNew(&viewer->m_port, 0, 0);
  if (err)
    return err;
  viewer->pData = hostData;
  //viewer->m_bFlow = true;
  viewer->m_dimension.width = 600;
  viewer->m_dimension.height = 800; 
  viewer->m_scale = 0;
  viewer->m_fontSize = 25;
  viewer->pszFontname = FskStrDoCopy("Times New Roman");
  viewer->fileContents = (char*)FskMemPtrAlloc(hostData->nFilesize+1);
  hostData->fileReadProc(hostData->fileReadProcData, hostData->nFilesize, viewer->fileContents, &bytesRead, 0);
  viewer->fileContents[hostData->nFilesize]=0;
  viewer->fileSize = hostData->nFilesize;
  viewer->m_bDidParas = false;
  viewer->m_bNeedToPaginate = true;
  viewer->m_arrParas = NULL;
  viewer->m_arrLines = NULL;
  viewer->m_curPage = 0;
  viewer->m_totalPages = 0;
  return 0;
}

void setLinesPerPage(myViewer *viewer, int linesPerPage)
{
  printf("[testviewer]: setLinesPerPage: %d\n", linesPerPage);
  //if (linesPerPage!=viewer->m_linesPerPage)
  {
    viewer->m_linesPerPage = linesPerPage;
    viewer->m_totalPages = (arrayGetSize(viewer->m_arrLines)-1 + linesPerPage-1)/linesPerPage;
    printf("[testviewer]: total pages: %d\n", viewer->m_totalPages);
  }
}

int setFont(myViewer *viewer)
{
  FskFontInfo fontInfo;
  int err;
  printf("[testviewer]: setFont %s@%d\n", viewer->pszFontname, viewer->m_fontSize);
  err = FskPortSetTextFont(viewer->m_port, viewer->pszFontname);
  if (err)
    return err;  
  err = FskPortSetTextSize(viewer->m_port, viewer->m_fontSize);
  if (err)
    return err;
  err = FskPortSetTextStyle(viewer->m_port, TS_PLAIN);
  if (err)
    return err;
  FskPortGetFontInfo(viewer->m_port, &fontInfo);
  viewer->m_lineHeight = fontInfo.tmAscent+fontInfo.tmDescent+((fontInfo.tmExternalLeading>0)?fontInfo.tmExternalLeading:-fontInfo.tmExternalLeading);
  printf("[testviewer]: fontInfo %d,%d,%d\n", fontInfo.tmAscent, fontInfo.tmDescent, fontInfo.tmExternalLeading);
  return 0;
}

void Paginate(myViewer* viewer)
{
  int i, k, parStart, lineStart, parLen, endpos;
  if (!viewer->m_bDidParas)
  {
    printf("[testviewer] Detecting paragraphs\n");
    arrayDispose(viewer->m_arrParas);
    viewer->m_arrParas = arrayNew(5);
    i = 0;
    parStart = 0;
    while (i<viewer->fileSize)
    {
      if (viewer->fileContents[i]=='\r' || viewer->fileContents[i]=='\n')
      {
        if (viewer->fileContents[i]=='\r' && viewer->fileContents[i+1]=='\n')
          i+=2;
        else
          i++;
        //printf("%d ",parStart);
        arrayAppendItem(viewer->m_arrParas, parStart);
        parStart = i;
      }
	    else
	      i++;
    }
    //printf("\n");
    arrayAppendItem(viewer->m_arrParas, viewer->fileSize);
    printf("[testviewer] total paragraphs: %d\n",arrayGetSize(viewer->m_arrParas)-1);
    viewer->m_bDidParas = true;
  }
  if (viewer->m_bNeedToPaginate)
  {
    printf("[testviewer] doing pagination\n");  
    arrayDispose(viewer->m_arrLines);
    viewer->m_arrLines = arrayNew(10);
    for (i=0;i<arrayGetSize(viewer->m_arrParas)-1;i++)
    {
      lineStart = arrayGetItem(viewer->m_arrParas, i);
      parLen = arrayGetItem(viewer->m_arrParas, i+1) - lineStart;
      //printf("parlen1: %d ",parLen);
      if (parLen && (viewer->fileContents[lineStart+parLen-1]=='\r' || viewer->fileContents[lineStart+parLen-1]=='\n')) parLen--;
      if (parLen && (viewer->fileContents[lineStart+parLen-1]=='\r' || viewer->fileContents[lineStart+parLen-1]=='\n')) parLen--;
      k = 0;
      //printf("parlen2: %d ",parLen);
      if (parLen<=0) {
        arrayAppendItem(viewer->m_arrLines, lineStart);
        //printf("%d ",lineStart);
      }
      while (k < parLen)
      {
        endpos=0;
        FskPortTextFitWidth(viewer->m_port, viewer->fileContents+lineStart+k, parLen-k, 
                            viewer->m_dimension.width, FW_WORDS, &endpos, NULL);
        arrayAppendItem(viewer->m_arrLines, lineStart+k);
        //printf("%d ",lineStart+k);
        k += endpos;
      }
      //printf("\n");
    }
    arrayAppendItem(viewer->m_arrLines, viewer->fileSize);
    setLinesPerPage(viewer, viewer->m_dimension.height/viewer->m_lineHeight);
    printf("[testviewer] total lines: %d\n",arrayGetSize(viewer->m_arrLines)-1);
    printf("[testviewer] finished pagination\n");
    viewer->m_bNeedToPaginate = false;
  }
}

int myDispose(ViewerData* hostData, myViewer* viewer)
{
  printf("[testviewer] Dispose\n");
  arrayDispose(viewer->m_arrParas);
  arrayDispose(viewer->m_arrLines);
  FskPortDispose(viewer->m_port);
  FskMemPtrDispose(viewer->fileContents);
  FSK_DELETE(viewer);
  return 0;
}

FskColor clWhite = {0xFF,0xFF,0xFF,0xFF};
FskColor clBlack = {0,0,0,0xFF};

int myRender(ViewerData* hostData, myViewer* viewer, FskBitmap** ppBitmap) //0=success
{
  FskRectangle rDim;
  FskRectangle rCrop;
  FskBitmap* bmp;
  int err,i,nLines,lineHeight,maxY,lineStart,lineLen;

  printf("[testviewer] Render\n");
  //__asm _emit 0xCC;
  *ppBitmap = 0;
  FskRectangleSet(&rDim, 0, 0, viewer->m_dimension.width, viewer->m_dimension.height);
  FskRectangleIntersect(&viewer->m_crop, &rDim, &rCrop);
  err = FskBitmapNew(rCrop.width, rCrop.height, 1, &bmp);
  if (err)
    return err;
  err = FskPortSetBitmap(viewer->m_port, bmp);
  if (err)
  {
    FskBitmapDispose(bmp);
    return err;
  }
  FskPortSetOrigin(viewer->m_port, -rCrop.x, -rCrop.y);
  FskPortSetPenColor(viewer->m_port, &clWhite);
  FskPortRectangleFill(viewer->m_port, &rDim);
  FskPortSetPenColor(viewer->m_port, &clBlack);
  FskPortSetTextAlignment(viewer->m_port, TA_LEFT, TA_TOP);
  Paginate(viewer);
  maxY = rCrop.height;
  nLines = arrayGetSize(viewer->m_arrLines)-1;
  rCrop.height = viewer->m_lineHeight;
  rCrop.y = 0;
  for (i=viewer->m_curPage*viewer->m_linesPerPage;i<nLines && rCrop.y<maxY; i++, rCrop.y+=viewer->m_lineHeight)
  {
    lineStart = arrayGetItem(viewer->m_arrLines, i);
    lineLen = arrayGetItem(viewer->m_arrLines, i+1) - lineStart;
    if (lineLen && (viewer->fileContents[lineStart+lineLen-1]=='\r' || viewer->fileContents[lineStart+lineLen-1]=='\n')) lineLen--;
    if (lineLen && (viewer->fileContents[lineStart+lineLen-1]=='\r' || viewer->fileContents[lineStart+lineLen-1]=='\n')) lineLen--;
    if (lineLen) 
      FskPortTextDraw(viewer->m_port, viewer->fileContents+lineStart, lineLen, &rCrop);
  }
  
  FskPortSetBitmap(viewer->m_port, 0);
  *ppBitmap = bmp;
  return 0;
}

int myGetDimensions(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] GetDimensions %d, %d, prop type %d\n", unknown1, propid, prop->type);
  prop->dimValue = viewer->m_dimension;
  prop->type = PropType_Dimensions;
  printf("%dx%d\n", prop->dimValue.width, prop->dimValue.height);
  return 0;
}

int mySetDimensions(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] SetDimensions: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  if (viewer->m_dimension.width != prop->dimValue.width || viewer->m_dimension.height != prop->dimValue.height)
  {
    viewer->m_dimension = prop->dimValue;
    viewer->m_bNeedToPaginate = true;
  }
  printf("%dx%d\n", prop->dimValue.width, prop->dimValue.height);
  return 0;
}

int myGetSourceDimensions(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] GetSourceDimensions %d, %d, prop type %d\n", unknown1, propid, prop->type);
  prop->dimValue = viewer->m_srcDimension;
  prop->type = PropType_Dimensions;
  printf("%dx%d\n", prop->dimValue.width, prop->dimValue.height);
  return 0;
}

int mySetSourceDimensions(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{ 
  printf("[testviewer] SetSourceDimensions: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  if (viewer->m_srcDimension.width != prop->dimValue.width || viewer->m_srcDimension.height != prop->dimValue.height)
  {
    viewer->m_srcDimension = prop->dimValue;
    viewer->m_bNeedToPaginate = true;
  }
  printf("%dx%d\n", prop->dimValue.width, prop->dimValue.height);
  return 0;
}

int myGetFlow(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] GetFlow: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  prop->type = PropType_Boolean;
  prop->bValue = true;//viewer->m_bFlow;
  printf("%d\n", prop->bValue);
  return 0;
}

int mySetFlow(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] SetFlow: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  //viewer->m_bFlow = prop->bValue;
  printf("%d\n", prop->bValue);
  return 0;
}

int myGetPages(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] GetPages: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  prop->type = PropType_Integer;
  Paginate(viewer);
  prop->iValue = viewer->m_totalPages;
  printf("%d\n", prop->iValue);
  return 0;
}

int myGetPage(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] GetPage: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  prop->type = PropType_Integer;
  prop->iValue = viewer->m_curPage;
  printf("%d\n", prop->iValue);
  return 0;
}

int mySetPage(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  int page = prop->iValue;
  printf("[testviewer] SetPage: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  printf("%d\n", prop->iValue);
  Paginate(viewer);
  if (page<0) page = 0;
  if (page>viewer->m_totalPages) page = viewer->m_totalPages-1;
  if (page != viewer->m_curPage)
  {
    viewer->m_curPage = page;
  }
  return 0;
}

int myGetCrop(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{  
  printf("[testviewer] GetCrop: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  prop->type = PropType_Rect;
  prop->rectValue = viewer->m_crop;
  printf("%d,%d, %dx%d\n", prop->rectValue.x, prop->rectValue.y, prop->rectValue.width, prop->rectValue.height);
  return 0;
}

int mySetCrop(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{  
  printf("[testviewer] SetCrop: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  viewer->m_crop = prop->rectValue;
  printf("%d,%d, %dx%d\n", prop->rectValue.x, prop->rectValue.y, prop->rectValue.width, prop->rectValue.height);
  return 0;
}

int myGetTextScale(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{  
  printf("[testviewer] GetTextScale: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  prop->type = PropType_Integer;
  prop->iValue = viewer->m_scale;
  printf("%d\n", prop->iValue);
  return 0;
}

int fontSizes[] = {25,32,38};
int mySetTextScale(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{  
  int err, scale = prop->iValue;
  printf("[testviewer] SetTextScale: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  if (scale<0) scale=0;
  if (scale>2) scale=2;
  printf("%d -> %d\n", prop->iValue, scale);  
  if (viewer->m_scale != scale)
  {
    viewer->m_scale = scale;
    viewer->m_fontSize = fontSizes[scale];
    setFont(viewer);
    viewer->m_bNeedToPaginate = true;
  }
  return 0;
}

int mySetFont(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  int err;
  printf("[testviewer] SetFont: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  printf("%s\n",prop->strValue);
  if (FskStrCompare(viewer->pszFontname, prop->strValue))
  {
    FskMemPtrDispose(viewer->pszFontname);
    viewer->pszFontname = FskStrDoCopy(prop->strValue);
    setFont(viewer);
    viewer->m_bNeedToPaginate = true;
  }
  return 0;
}

int myGetFont(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  prop->type = PropType_String;
  prop->strValue = FskStrDoCopy(viewer->pszFontname);
  printf("[testviewer] GetFont: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  printf("%s\n",prop->strValue);
  return 0;
}

int mySetTextEngine(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  FskPort* port;
  int err;  
  printf("[testviewer] SetTextEngine: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  printf("%s\n",prop->strValue);
  err = FskPortNew(&port, 0, prop->strValue);
  if (err)
    return err;
  FskPortDispose(viewer->m_port);
  viewer->m_port = port;
  setFont(viewer);
  viewer->m_bNeedToPaginate = true;
  return 0;
}

int myGetLayoutCache(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  //prop->type = PropType_String;
  //prop->strValue = FskStrDoCopy(viewer->pszFontname);
  printf("[testviewer] GetLayoutCache: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  //printf("%s\n",prop->strValue);
  return 0;
}

int mySetLayoutCache(myViewer *viewer, int unknown1, int propid, PropertyValue* prop)
{
  printf("[testviewer] SetLayoutCache: %08X, prop id %d, prop type %d\n", unknown1, propid, prop->type);
  //printf("%s\n",prop->strValue);
  return 0;
}

int myMarkNewFromLocation(ViewerData* hostData, myViewer *viewer, void **pMark, int pageNo, FskPoint *point, char flag)
{
  int lineNo, err, nLines, lineLen, lineStart;
  int* mark;
  
  printf("[testviewer] MarkNewFromLocation: page %d, point %dx%d, flag %d\n", pageNo, point->x, point->y, flag);
  Paginate(viewer);
  if (pageNo<0 || pageNo>=viewer->m_totalPages)
  	return -7;
  nLines = arrayGetSize(viewer->m_arrLines)-1;
  if (err=FskMemPtrNewClear(sizeof(int), &mark))
    return err;
  *pMark = mark;
  if (flag & (MARK_PAGESTART|MARK_PAGEEND))
  { 
    if (flag & MARK_PAGESTART)
      lineNo = pageNo*viewer->m_linesPerPage;
    else
      lineNo = (pageNo+1)*viewer->m_linesPerPage;
    if (lineNo>nLines) lineNo=nLines;
    *mark = arrayGetItem(viewer->m_arrLines, lineNo);
  }
  else
  {
    lineNo = point->y/viewer->m_lineHeight;
    if (lineNo>=viewer->m_linesPerPage)
      lineNo = viewer->m_linesPerPage-1;
    lineNo += pageNo*viewer->m_linesPerPage;
    if (lineNo>nLines) {
      *mark = arrayGetItem(viewer->m_arrLines, nLines);
    }
    else
    {
      lineStart = arrayGetItem(viewer->m_arrLines, lineNo);
      lineLen = arrayGetItem(viewer->m_arrLines, lineNo+1) - lineStart;
      if (lineLen && (viewer->fileContents[lineStart+lineLen-1]=='\r' || viewer->fileContents[lineStart+lineLen-1]=='\n')) lineLen--;
      if (lineLen && (viewer->fileContents[lineStart+lineLen-1]=='\r' || viewer->fileContents[lineStart+lineLen-1]=='\n')) lineLen--;

      FskPortTextFitWidth(viewer->m_port, viewer->fileContents+lineStart, lineLen, point->x, 0, mark, NULL);
      *mark += lineStart;
    }
  }
  printf("  text pos %d\n", *mark);
  return 0;
}

int myMarkDispose(ViewerData* hostData, myViewer *viewer, void *mark)
{
  FskMemPtrDispose(mark);
  return 0;
}

int myMarkGetLocation(ViewerData* hostData, myViewer *viewer, int *mark, int *pageNo, FskPoint *point, int *part)
{
  int nLines, iLine, off, page, lineStart, high, low;
  FskRectangle rect;
  
  printf("[testviewer] MarkGetLocation: pos %d\n", *mark);

  Paginate(viewer);
  if (part) *part=0;
  nLines = arrayGetSize(viewer->m_arrLines)-1;
  off = *mark;
  if (off>arrayGetItem(viewer->m_arrLines, nLines))
    return -5;
  low = 0; high = nLines;
  while (low<high)
  {
    iLine = (low+high)/2;
    if (off<arrayGetItem(viewer->m_arrLines, iLine))
      high = iLine;
    else if (off>=arrayGetItem(viewer->m_arrLines, iLine+1))
      low = iLine;
    else 
      break;
  }
  page = iLine/viewer->m_linesPerPage;
  if (page >= viewer->m_totalPages)
    page = viewer->m_totalPages-1;
  if (pageNo) *pageNo = page;
  printf("  page %d", page);
  if (point)
  {
     point->y = viewer->m_lineHeight*(iLine - page*viewer->m_linesPerPage);
     lineStart = arrayGetItem(viewer->m_arrLines, iLine);
     FskPortTextGetBounds(viewer->m_port, viewer->fileContents+lineStart, off - lineStart, &rect);
     point->x = rect.width;
     printf("  point %dx%d\n", point->x, point->y);
  }
  return 0;
}

int myMarkCompare(ViewerData* hostData, myViewer *viewer, int* mark1, int* mark2, int* result)
{
  if (*mark1==*mark2)
    *result=0;
  else
    *result=(*mark1>*mark2)?1:-1;
  return 0;
}

PropertyDesc myPropTable[] =  {
  {DocViewerProp_dimensions, PropType_Dimensions, &myGetDimensions, &mySetDimensions},
  {DocViewerProp_pages, PropType_Integer, &myGetPages, 0},
  {DocViewerProp_page, PropType_Integer, &myGetPage, &mySetPage},
  {DocViewerProp_crop, PropType_Rect, &myGetCrop, &mySetCrop}, 
  {DocViewerProp_textScale, PropType_Integer, &myGetTextScale, &mySetTextScale}, 
  {DocViewerProp_flow, PropType_Boolean, &myGetFlow, 0/*&mySetFlow*/}, 
  //{DocViewerProp_layoutCache, PropType_Pointer, &myGetLayoutCache, &mySetLayoutCache}, 
  {DocViewerProp_sourceDimensions, PropType_Dimensions, &myGetSourceDimensions,  &mySetSourceDimensions},
  {DocViewerProp_font, PropType_String, &myGetFont, &mySetFont}, 
  //{DocViewerProp_onProgress, PropType_Method, 0, &mySetProgress},
  {DocViewerProp_textEngine, PropType_String, 0,  &mySetTextEngine},
  {0}
};

static DocViewer myDocViewer = {
 /* 0*/ &myCanHandle,
 /* 4*/ &myNew,
 /* 8*/ &myDispose,
 /* C*/ &myRender,
 /*10*/ 0, // Find
 /*14*/ 0, // GetMetadata;
 /*18*/ &myMarkNewFromLocation,
 /*1C*/ &myMarkDispose,
 /*20*/ &myMarkGetLocation,
 /*24*/ &myMarkCompare,
 /*28*/ 0, // SpanGetBounds;
 /*2C*/ 0, // SpanGetText;
 /*30*/ myPropTable, // PropTable;
 /*34*/ 0, // MarkSerialize;
 /*38*/ 0, // MarkParse;
};

int __declspec(dllexport) fskLoad()
{
	printf("[testviewer] fskLoad()\n");
  FskExtensionInstall(EXTENSION_DOCVIEWER, &myDocViewer);
  return 0;
}

void __declspec(dllexport) fskUnload()
{
	printf("[testviewer] fskUnload()\n");
	FskExtensionUninstall(EXTENSION_DOCVIEWER, &myDocViewer);
}

//#pragma comment(linker, "/export:fskLoad")
//#pragma comment(linker, "/export:fskUnload")
